/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file daisylibao.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 15.02.2006
 *
 * An interface against libao
 */

 
/**
 * includes
 */
#include <ao/ao.h>
#include <stdio.h>
#include <stdbool.h>
#include <errno.h>

#include "daisylibao.h"
#include "common.h"
#include "audio.h"
#include "report.h"
#include "snprintf/snprintf.h"


/*
 * variables
 */ 
static ao_device *device;

/**
 * declarations
 */
static
void ao_open_live_error (void);



/**
 * This will output samples from a buffer. It basically wraps the ao_play() function.
 */
int libao_play (void *output_samples, int num_bytes)
{
	return ao_play(device, output_samples, num_bytes);
}

/**
 * This will initialize libao for us.
 * true on success, false on fail.
 */
bool libao_init (int byte_format, int channels, int samplerate)
{
	ao_sample_format format;
	int libao_driver_id = -1;
	ao_initialize ();
	libao_driver_id =  ao_default_driver_id ();
	if (libao_driver_id == -1) {
		report("Could not get a valid driver ID", REP_ERROR);
		return false;
	}
	format.bits = 16; /** Do decoder always give us 16bits? */
/**	format.byte_format = AO_FMT_LITTLE; */
	format.byte_format = byte_format;
	format.channels = channels;
	format.rate = samplerate;

	device = ao_open_live (libao_driver_id, &format, NULL);
	if (device == NULL)
	{
		ao_open_live_error();
		return false;
	}
	else
	{
		if (DEBUG_LIBAO)
		{
			char msg[STRLEN];
			report ("libao initialized", REP_DEBUG);
			snprintf (msg, STRLEN, "libao info:\n\tbyte_format: %d\n\tchannels: %d\n\tsamplerate: %d", byte_format, channels, samplerate);
			report (msg, REP_DEBUG);
		}
	}
	return true;
}


/**
 * This function is called if ao_open_live returns NULL. This means we have
 * an error, and this method reports it based on errno
 */
static
void ao_open_live_error ()
{
	switch (errno)
	{
		case AO_ENODRIVER:
			report ("No driver corresponts to driver ID.", REP_ERROR);
			break;
		case AO_ENOTLIVE:
			report ("This driver is not a live output device.", REP_ERROR);
			break;
		case AO_EBADOPTION:
			report ("A valid option key has an invalid value.", REP_ERROR);
			break;
		case AO_EOPENDEVICE:
			report ("Cannot open the device.", REP_ERROR);
			break;
		case AO_EFAIL:
			report ("Unknown reason.", REP_ERROR);
			break;
		default:
			report ("ao_open_live_error called with unknown error", REP_ERROR);
			break;
	}
}


/**
 * Checks if device is NULL (uninitialized)
 */
bool ao_device_is_null (void)
{
	return (int) device;
}


/*
 * This will close libao
 */
void libao_close (void)
{
	ao_close (device);
	ao_shutdown ();
	if (DEBUG_LIBAO) report ("libao closed", REP_DEBUG);
}
