/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file smilparser.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 23.02.2006
 * 
 */
 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml2/libxml/xmlreader.h>

#include "ncxsmilparser.h"
#include "control.h"
#include "report.h"
#include "common.h"
#include "snprintf/snprintf.h"

#ifdef LIBXML_READER_ENABLED

/* static function declarations */
static int parseSmilHead (xmlTextReaderPtr reader);
static int parseSmilLayout (xmlTextReaderPtr reader);
static int parseSmilMeta (xmlTextReaderPtr reader);
static int parseSmilRegion (xmlTextReaderPtr reader);
static int parseSmilCustomAttributes (xmlTextReaderPtr reader);
static int parseSmilCustomTest (xmlTextReaderPtr reader);
static int parseSmilBody (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseSmilSeq (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseSmilPar (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseSmilText (xmlTextReaderPtr reader, struct Node *tmpNode);
static int parseSmilAudio (xmlTextReaderPtr reader, struct Node *tmpNode);
static int parseSmilImg (xmlTextReaderPtr reader, struct Node *tmpNode);
static int parseSmilA (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseXML (char* filename, char* fragment, struct Node *tmpNode);
static int parseXMLHead (xmlTextReaderPtr reader);
static int parseXMLTitle (xmlTextReaderPtr reader);
static int parseXMLMeta (xmlTextReaderPtr reader);
static int parseXMLFrontmatter (xmlTextReaderPtr reader);
static int parseXMLBodymatter (xmlTextReaderPtr reader, char* fragment, struct Node *tmpNode);
static int parseXMLRearmatter (xmlTextReaderPtr reader);
static int parseXMLDoctitle (xmlTextReaderPtr reader);
static int parseXMLDocauthor (xmlTextReaderPtr reader);
static int parseXMLLevel (xmlTextReaderPtr reader);
static int parseXMLLevel1 (xmlTextReaderPtr reader, char* fragment, struct Node *tmpNode);
static int parseXMLBook (xmlTextReaderPtr reader, char *fragment, struct Node *tmpNode);
static int parseXMLCovertitle (xmlTextReaderPtr reader);
/*
static int parseXMLPagenum (xmlTextReaderPtr reader);
static int parseXMLH1 (xmlTextReaderPtr reader);
static int parseXMLLevel2 (xmlTextReaderPtr reader);
static int parseXMLBridgehead (xmlTextReaderPtr reader);
static int parseXMLP (xmlTextReaderPtr reader);
static int parseXMLList (xmlTextReaderPtr reader);
static int parseXMLDl (xmlTextReaderPtr reader);
static int parseXMLDiv (xmlTextReaderPtr reader);
static int parseXMLBlockquote (xmlTextReaderPtr reader);
static int parseXMLImg (xmlTextReaderPtr reader);
static int parseXMLImggroup (xmlTextReaderPtr reader);
static int parseXMLPoem (xmlTextReaderPtr reader);
static int parseXMLLinegroup (xmlTextReaderPtr reader);
static int parseXMLByline (xmlTextReaderPtr reader);
static int parseXMLDateline (xmlTextReaderPtr reader);
static int parseXMLEpigraph (xmlTextReaderPtr reader);
static int parseXMLTable (xmlTextReaderPtr reader);
static int parseXMLAddress (xmlTextReaderPtr reader);
static int parseXMLLine (xmlTextReaderPtr reader);
static int parseXMLAuthor (xmlTextReaderPtr reader);
static int parseXMLProdnote (xmlTextReaderPtr reader);
static int parseXMLSidebar (xmlTextReaderPtr reader);
static int parseXMLNote (xmlTextReaderPtr reader);
static int parseXMLA (xmlTextReaderPtr reader);
tatic int parseXMLCite (xmlTextReaderPtr reader);
static int parseXMLSamp (xmlTextReaderPtr reader);
static int parseXMLKbd (xmlTextReaderPtr reader);
static char* parseXMLHtmltag (xmlTextReaderPtr reader);
*/

/* global variables*/
static int alreadyfound = 0; /* Used in the smil-parsing */


/************************ Code for parsing a *.smil files ************************/


/**
 * Method for parsing of a *.smil file
 * 
 * @param tmpSmilNode - the smilNode for the smilfile to parse.
 * @returns 1 in case of success and -1 in case of error
 */
int parseNCXSmil (struct SmilNode *tmpSmilNode)
{	
 	const xmlChar *name = NULL;
 	xmlTextReaderPtr reader = NULL;
 	int ret = 1, type = 1;
 	char tmp[STRLEN];
 	alreadyfound = 0;
 	
	snprintf (tmp, sizeof (tmp), "%s%s", __path,tmpSmilNode->anchor); 
 	/* open the smil file */
 	reader = xmlReaderForFile (tmp, NULL, 0);
 	if (reader == NULL)
 	{
 		char error[STRLEN];
 		snprintf (error, sizeof (error), "Failed to open SMIL file: %s", tmpSmilNode->anchor);
 		report (error, REP_WARNING);
 		return -1;
 	}
 	else 
	{
		if (DEBUG_NCXSMILPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing SMIL file: %s", tmpSmilNode->anchor);
	 		report (error, REP_DEBUG);
		}
 			
	 	/* read the first tag */
	 	ret = xmlTextReaderRead(reader);
	 	if (ret != 1)
	 	{
	 		char error[STRLEN];
	 		snprintf (error, sizeof (error), "Failed to parse SMIL file: %s", tmpSmilNode->anchor);
 			report (error, REP_WARNING);
 			return -1;
	 	}
	 	name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;	
		if (!strcmp ((const char *)name, "smil")) 
		{
			ret = xmlTextReaderRead(reader);
	 		if (ret != 1)
		 	{
		 		char error[STRLEN];
		 		snprintf (error, sizeof (error), "Failed to parse SMIL file: %s", tmpSmilNode->anchor);
	 			report (error, REP_WARNING);
	 			return -1;
		 	}
		 	name = xmlTextReaderConstName (reader);
			if (name == NULL) return 0;
			type = xmlTextReaderNodeType(reader);
		}
	 	/* while we haven't reached the smil endtag */
	 	while (type != 15 && strcmp ((const char *)name, "smil"))
	 	{
			/* Attributes:
	 		 * id
	 		 * class
	 		 * title
			 * xmlns
			 * xml:lang
			 */
			 
			/* parse tag if <head> */
			if (!strcmp ((const char *)name, "head"))
			{ 
				ret = parseSmilHead (reader);
				if (ret != 1)
				{
					char error[STRLEN];
			 		snprintf (error, sizeof (error), "Failed to parse SMIL file: %s", tmpSmilNode->anchor);
		 			report (error, REP_WARNING);
		 			return -1;
			 	}			 	
			}
			/* parse tag if <body> */
			else if (!strcmp ((const char *)name, "body"))
			{ 
				ret = parseSmilBody (reader, tmpSmilNode);
				if (ret != 1)
				{
					char error[STRLEN];
			 		snprintf (error, sizeof (error), "Failed to parse SMIL file: %s", tmpSmilNode->anchor);
		 			report (error, REP_WARNING);
		 			return -1;
			 	}			 	
			}
			ret = xmlTextReaderRead (reader);
			if (ret != 1) return ret;
			name = xmlTextReaderConstName (reader);
			if (name == NULL) return 0;		
			type = xmlTextReaderNodeType(reader);
		}
		xmlFreeTextReader (reader);
		if (ret != 1)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Failed to parse SMIL file: %s", tmpSmilNode->anchor);
 			report (error, REP_WARNING);
 			return -1;
		}
	}
	return ret;
}


/**
 * Parse Head-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilHead(xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret, type;
	ret = type = 1;
	
	/* Attributes:
	 * id
	 * class
	 * title
	 * xml:lang
	 */
	
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <HEAD>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing SMIL <HEAD>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	/* while we haven't reached the head-close tag */
	while (strcmp ((const char *)name, "head"))
	{	
		/* parse tag if <layout> ? */
		if (!strcmp ((const char *)name, "layout")) ret = parseSmilLayout (reader);
		/* parse tag if <meta> * */
		else if (!strcmp ((const char *)name, "meta")) ret = parseSmilMeta (reader);
		/* parse tag if <customAttributes> ? */
		else if (!strcmp ((const char *)name, "customAttributes")) ret = parseSmilCustomAttributes (reader);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing SMIL <HEAD>, endtag expected", REP_ERROR);
		return -1;
	}		
	return ret;
}

/**
 * Parse Layout-tag. Currently not implemented.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilLayout (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret, type;
	ret = type = 1;
	
	/* Attributes:
	 * id
	 * class
	 * title
	 * xml:lang
	 */
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <LAYOUT>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing SMIL <LAYOUT>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	/* while we haven't reached the layout-close tag */
	while (strcmp ((const char *)name, "layout"))
	{	
		/* parse tag if <region> + */
		if (!strcmp ((const char *)name, "region")) ret = parseSmilRegion (reader);		
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}	
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing SMIL <LAYOUT>, endtag expected", REP_ERROR);
		return -1;
	}
	return ret;
}

/**
 * Parse Meta-tag. Currently not implemented.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilMeta (xmlTextReaderPtr reader)
{
	int ret = 1;
	
	/* Attributes:
	 * content
	 * name
	 */
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <META>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;	
}

/**
 * Parse Region-tag. Currently not implemented
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilRegion (xmlTextReaderPtr reader)
{
	int ret = 1;
	
	/* Attributes:
	 * id
	 * bottom
	 * left
	 * right
	 * top
	 * height
	 * width
	 * fit
	 * backgroundColor
	 * showBackground
	 * z-index
	 * xml:lang
	 */
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <REGION>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;	
}


/**
 * Parse CustomAttributes-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilCustomAttributes (xmlTextReaderPtr reader)
{	
	const xmlChar *name = NULL;
	int ret, type;
	ret = type = 1;
	
	/* Attributes:
	 * id
	 * class
	 * title
	 * xml:lang
	 */
	 
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <CUSTOMATTRIBUTES>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing SMIL <CUSTOMATTRIBUTES>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	/* while we haven't reached the customattributes-close tag */
	while (strcmp ((const char *)name, "customAttributes"))
	{	
		/* parse tag if <customTest> + */
		if (!strcmp ((const char *)name, "customTest")) ret = parseSmilCustomTest (reader);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing SMIL <CUSTOMATTRIBUTES>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parse CustomTest-tag. Currently not implemented
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilCustomTest (xmlTextReaderPtr reader)
{	
	int ret;
	/* Attributes:
	 * id
	 * class
	 * title
	 * xml:lang
	 * defaultState
	 * override
	 */

	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <CUSTOMTEST>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;	
}


/**
 * Parse Body-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilBody (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	const xmlChar *name = NULL;
	int ret, type;
	ret = type = 1;

	/* Attributes:
	 * id
	 * class
	 * title
	 * xml:lang
	 */

	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <BODY>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing SMIL <BODY>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	/* while we haven't reached that body close tag */
	while (strcmp ((const char *)name, "body"))
	{	
		/* parse tag if <par> + */
		if (!strcmp ((const char *)name, "par")) ret = parseSmilPar (reader, tmpSmilNode);
		/* parse tag if <seq> + */
		else if (!strcmp ((const char *)name, "seq")) ret = parseSmilSeq (reader, tmpSmilNode);
		/* parse tag if <text> + */
		else if (!strcmp ((const char *)name, "text")) ret = parseSmilText (reader, NULL);
		/* parse tag if <audio> + */
		else if (!strcmp ((const char *)name, "audio")) ret = parseSmilAudio (reader, NULL);
		/* parse tag if <img> + */
		else if (!strcmp ((const char *)name, "img")) ret = parseSmilImg (reader, NULL);
		/* parse tag if <a> + */
		else if (!strcmp ((const char *)name, "a")) ret = parseSmilA (reader, tmpSmilNode);
			
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1)  return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing SMIL <BODY>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}

/**
 * Parse Main-Seq-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilSeq (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	const xmlChar *name = NULL;
	int ret, type;
	ret = type = 1;

	/* Attributes:
	 * id
	 * class
	 * customTest
	 * dur
	 * xml:lang
	 * end
	 * fill
	 */
	 
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <SEQ>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing SMIL <SEQ>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	/* while we haven't reached the seq endtag */
	while (type != 15 && strcmp ((const char *)name, "seq"))
	{	
		/* parse tag if <par> + */
		if (!strcmp ((const char *)name, "par")) ret = parseSmilPar (reader, tmpSmilNode);
		/* parse tag if <seq> + */
		else if (!strcmp ((const char *)name, "seq")) ret = parseSmilSeq (reader, tmpSmilNode);
		/* parse tag if <text> + */
		else if (!strcmp ((const char *)name, "text")) ret = parseSmilText (reader, NULL);
		/* parse tag if <audio> + */
		else if (!strcmp ((const char *)name, "audio")) ret = parseSmilAudio (reader, NULL);
		/* parse tag if <img> + */
		else if (!strcmp ((const char *)name, "img")) ret = parseSmilImg (reader, NULL);
		/* parse tag if <a> + */
		else if (!strcmp ((const char *)name, "a")) ret = parseSmilA (reader, tmpSmilNode);
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;		
		type = xmlTextReaderNodeType(reader);
	}	
	return ret;
}

/**
 * Parse Par-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilPar (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	const xmlChar *name = NULL;
	xmlChar *tmp = NULL;
	int ret = 1, type = 1;
	
	/* Attributes:
	 * id
	 * class
	 * customTest
	 * xml:lang
	 */

	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <PAR>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
		{   if (DEBUG_NCXPARSER) report ("Failed parsing SMIL <PAR>", REP_ERROR);
			return -1;
		}
	if(alreadyfound == 0)
	{
		/* go to the fragment identifier point and parses from there. */
		tmp = xmlTextReaderGetAttribute (reader, (xmlChar *)"id");		
		/* if the par tag has an id attribute, read it and check wheter it matches the fragmentid.
		   if there is none id attribute, set alreadyfound to 1 and parse the whole smilfile */
		if (tmp != NULL) 
		{	
			if (tmpSmilNode->fragmentIdentifier != NULL)
			{	
				if (!strcmp ((char *)tmp, tmpSmilNode->fragmentIdentifier))
					alreadyfound = 1; 
			}
		}
		else
			alreadyfound = 1;
		/* should we continue parsing if the id tag is missing?
		   the way it is now, if the id tag is missing we parse from here and out  */
		if (tmp != NULL) xmlFree (tmp); tmp = NULL;
	}
	if (alreadyfound == 1)
	{
		struct Node *tmpNode;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
		
		/* adding a new node to the linked list*/
		tmpNode = addNewNode();
		
		/* while we haven't reached the par-end tag */
		while (strcmp ((const char *)name, "par"))
		{	
			/* parse tag if <seq> + */
			if (!strcmp ((const char *)name, "seq")) ret = parseSmilSeq (reader, tmpSmilNode);
			/* parse tag if <text> + */
			else if (!strcmp ((const char *)name, "text")) ret = parseSmilText (reader, tmpNode);
			/* parse tag if <audio> + */
			else if (!strcmp ((const char *)name, "audio")) ret = parseSmilAudio (reader, tmpNode);
			/* parse tag if <img> + */
			else if (!strcmp ((const char *)name, "img")) ret = parseSmilImg (reader, tmpNode);
			/* parse tag if <a> + */
			else if (!strcmp ((const char *)name, "a")) ret = parseSmilA (reader, tmpSmilNode);
			if (ret != 1) return ret;
			ret = xmlTextReaderRead (reader);
			if (ret != 1) return ret;
			name = xmlTextReaderConstName (reader);
			if (name == NULL) return 0;
		}
	}
	else 
	{	ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
		while (strcmp ((const char *)name, "par"))
		{	
			ret = xmlTextReaderRead (reader);
			if (ret != 1) return ret;
			name = xmlTextReaderConstName (reader);
			if (name == NULL) return 0;
		}	
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing SMIL <PAR>, endtag expected", REP_ERROR);
		return -1;
	}
	return ret;
}


/**
 * Parse Text-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilText (xmlTextReaderPtr reader, struct Node *tmpNode)
{
	xmlChar *attr = NULL;
	const xmlChar *name = NULL;
	char  *textsrc = NULL, *fragment = NULL, *tmp = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * src
	 * type
	 * region
	 * xml:lang
	 */
	 
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <TEXT>", REP_DEBUG);
	name = xmlTextReaderConstName (reader);
	if (!strcmp((const char *)name, "#text")) {
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
	}
	attr = xmlTextReaderGetAttribute(reader, (xmlChar *)"src");	
	
	/* splits the text on #.
	   get the anchor url  */
	textsrc = strtok ((char *)attr, "#");
	
	if (DEBUG_NCXSMILPARSER)
	{
		char error[STRLEN];
		snprintf (error, sizeof (error), "Parsing SMIL <TEXT> XMLFile: %s", textsrc);
		report (error, REP_DEBUG);
	}
	
	/* storing the filename of the text file */
	tmp = (char *) malloc (strlen (textsrc)+1);
	strcpy (tmp, textsrc);
	tmpNode->textfilename = tmp;
	
	/* get the fragment identifier */
	fragment = strtok (NULL, "#");
	
	if (DEBUG_NCXSMILPARSER)
	{
		char error[STRLEN];
		snprintf (error, sizeof (error), "Parsing SMIL <TEXT> Fragment Identifier: %s", fragment);
		report (error, REP_DEBUG);
	}
	
	/* storing the fragment identifier for the passage in the text file */
	tmp = (char *) malloc (strlen (fragment)+1);
	strcpy (tmp, fragment);
	tmpNode->fragmentIdentifier = tmp;
	
	/* parse text file */
	if (parseXML(textsrc, fragment, tmpNode) != 0)
	{
		if (DEBUG_NCXSMILPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Failed to parse XML file: %s", tmpNode->textfilename);
	 		report (error, REP_WARNING);
		}
 		return -1;
	}
	
	if (attr != NULL) xmlFree (attr);
	return ret;
}

/**
 * Parse Audio-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilAudio (xmlTextReaderPtr reader, struct Node *tmpNode)
{
	xmlChar *tmp = NULL;
	char *tempChar = NULL;
	int ret = 1;

	/* Attributes:
	 * id
	 * src
	 * type
	 * clipBegin
	 * clipEnd
	 * region
	 * xml:lang
	 */
	/* fetch the url anchor to the audio file */
	tmp = xmlTextReaderGetAttribute (reader, (xmlChar *)"src");
	if (tmp != NULL)
	{
		tempChar = (char *) malloc (strlen((char *)tmp)+1);
		strcpy(tempChar, (char *)tmp);
		tmpNode->audiofilename = tempChar;
		
		if (DEBUG_NCXSMILPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing SMIL <AUDIO> Audiofile: %s", tmp);
			report (error, REP_DEBUG);
		}
		
		if (tmp != NULL) xmlFree (tmp); tmp = NULL;
		
		/* fetch clipBegin */
		tmp = xmlTextReaderGetAttribute (reader, (xmlChar *)"clipBegin");
		if (tmp != NULL)
		{
			tempChar = getTime((char *)tmp);
			tmpNode->audiofilename = tempChar;
			
			if (DEBUG_NCXSMILPARSER)
			{
				char error[STRLEN];
				snprintf (error, sizeof (error), "Parsing SMIL <AUDIO> clipBegin: %s", tempChar);
				report (error, REP_DEBUG);
			}
			
			if (tmp != NULL) xmlFree (tmp); tmp = NULL;
			
			/* fetch clipEnd */
			tmp = xmlTextReaderGetAttribute (reader, (xmlChar *)"clipEnd");
			if (tmp != NULL)
			{
				tempChar = getTime((char *)tmp);
				tmpNode->audiofilename = tempChar;
				
				if (DEBUG_NCXSMILPARSER)
				{
					char error[STRLEN];
					snprintf (error, sizeof (error), "Parsing SMIL <AUDIO> clipEnd: %s", tempChar);
					report (error, REP_DEBUG);
				}
			}
		}
		else
		{
			ret = 0;
		}
	}
	else
	{
		ret = 0;	
	}
	if (tmp != NULL) xmlFree (tmp); tmp = NULL;	
	return ret;
}


/**
 * Parse Img-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilImg (xmlTextReaderPtr reader, struct Node *tmpNode)
{
	xmlChar *tmp = NULL;
	char *tempChar = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * src
	 * type
	 * region
	 * xml:lang
	 */
	 
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <IMG>", REP_DEBUG);
	/* fetch the url anchor to the audio file */
	tmp = xmlTextReaderGetAttribute (reader, (xmlChar *)"src");
	if (tmp != NULL)
	{
		tempChar = (char *) malloc (strlen((char *)tmp)+1);
		strcpy(tempChar, (char *)tmp);
		
		/* store the image in the node */
		tmpNode->image = tempChar;
	
		if (DEBUG_NCXSMILPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing SMIL <IMG> Imagefile: %s", tmp);
			report (error, REP_DEBUG);
		}	
	}
	else
	{
		ret = 0;	
	}
	if (tmp != NULL) xmlFree (tmp);	tmp = NULL;
	return ret;
}


/**
 * Parse A-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilA (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	const xmlChar *name = NULL;
	int ret, type;
	ret = type = 1;
	
	/* Attributes:
	 * id
	 * class
	 * title
	 * xml:lang
	 * href
	 * external
	 */
	 
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <A>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing SMIL <AT>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	/* while we haven't reached the a-close tag */
	while (strcmp ((const char *)name, "a"))
	{	
		/* parse tag if <par> * */
		if (!strcmp ((const char *)name, "par")) ret = parseSmilPar (reader, tmpSmilNode);
		/* parse tag if <seq> * */
		else if (!strcmp ((const char *)name, "seq")) ret = parseSmilSeq (reader, tmpSmilNode);
		/* parse tag if <text> * */
		else if (!strcmp ((const char *)name, "text")) ret = parseSmilText (reader, NULL);
		/* parse tag if <audio> * */
		else if (!strcmp ((const char *)name, "audio")) ret = parseSmilAudio (reader, NULL);
		/* parse tag if <img> * */
		else if (!strcmp ((const char *)name, "img")) ret = parseSmilImg (reader, NULL);		
		if (ret != 1) return ret;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing SMIL <A>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}
/************************ END of *.smil parser code ************************/

/********************** Code for parsing a XML files **********************/


/**
 * Parses a xml-file.
 * 
 * @param filename the filename to the xml-file.
 * @param fragment the fragment to fetch.
 */
static int parseXML (char* filename, char* fragment, struct Node *tmpNode)
{
	const xmlChar *name = NULL;
	int ret = 1;
 	xmlTextReaderPtr reader = NULL;	
	
	char tmp[STRLEN];
	snprintf (tmp, sizeof (tmp), "%s%s", __path, filename);

	/* open the xml file */
 	reader = xmlReaderForFile (tmp, NULL, 0); 
	if (reader == NULL)
 	{
 		char error[STRLEN];
 		snprintf (error, sizeof (error), "Failed to open XML file: %s", filename);
 		report (error, REP_WARNING);
 		return -1;
 	}
 	else 
	{
		if (DEBUG_NCXXML)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing XML file: %s", filename);
	 		report (error, REP_DEBUG);
		}
		
		ret = xmlTextReaderRead(reader);
		if (ret != 1)
	 	{
	 		char error[STRLEN];
	 		snprintf (error, sizeof (error), "Failed to parse XML file: %s", filename);
 			report (error, REP_WARNING);
 			return -1;
	 	}
		while(ret == 1)
		{	
			/* Attributes:
	 		 */
	 		 
			name = xmlTextReaderConstName (reader);	
			/* parse tag if <head> */
			if (!strcmp ((const char *)name, "head"))
			{
				ret = parseXMLHead (reader);
				if (ret != 1)
				{
					char error[STRLEN];
			 		snprintf (error, sizeof (error), "Failed to parse XML file: %s", filename);
		 			report (error, REP_WARNING);
		 			return -1;
			 	}
			}
			/* parse tag if <book> */
			else if (!strcmp ((const char *)name, "book"))
			{
				ret = parseXMLBook (reader, fragment, tmpNode);
				xmlFreeTextReader (reader);
				if (ret == 1) return 0;
				else return ret;
			}
			ret = xmlTextReaderRead (reader);
		}
		xmlFreeTextReader (reader);
		if (ret != 0)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Failed to parse XML file: %s", filename);
 			report (error, REP_WARNING);
 			return -1;
		}
	}
	return ret;
}

/**
 * Parse Head-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLHead (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret, type;
	ret = type = 1;
	
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <HEAD>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <HEAD>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	
	/* while we haven't reached the head-close tag */
	while (strcmp ((const char *)name, "head"))
	{	
		/* parse tag if <title> */
		if (!strcmp ((const char *)name, "title"))
		{
			ret = parseXMLTitle (reader);
			if (ret != 1) return ret;
		}
		/* parse tag if <meta> */
		else if (!strcmp ((const char *)name, "meta"))
		{
			ret = parseXMLMeta (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <HEAD>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}

/**
 * Parse Title-tag. Extracts the title.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLTitle (xmlTextReaderPtr reader)
{
	xmlChar *value = NULL;
	char *tmp = NULL;
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <TITLE>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	
	/* if the title tag has a value	 */
	if (xmlTextReaderHasValue (reader)){
		value = xmlTextReaderValue (reader);
		tmp =  removewhitespaces ((char *)value);
	/*	printf ("\t-%s-",tmp);
		TODO: The title is never used.. */
	}
	if(value != NULL) 
		xmlFree(value); value = NULL;
	if(tmp != NULL)
		free(tmp); tmp = NULL;
	ret = xmlTextReaderRead (reader);
	return ret;
}

/**
 * Parse Meta-tag. Currently not implemented.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLMeta (xmlTextReaderPtr reader){
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <META>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;	
}

/**
 * Parse Body-tag.
 *
 * @param reader XmlTextReader.
 * @param fragment the fragment to fetch.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLBook (xmlTextReaderPtr reader, char *fragment, struct Node *tmpNode)
{
	const xmlChar *name = NULL;	
	int ret, type;
	ret = type = 1;
	
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <BOOK>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <BOOK>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	
	/* while we haven't reached the book-close tag */
	while (strcmp ((const char *)name, "book"))
	{	
		/* parse tag if <frontmatter> */
		if (!strcmp ((const char *)name, "frontmatter"))
		{
			ret = parseXMLFrontmatter (reader);
			if (ret != 1) return ret;
		}
		/* parse tag if <bodymatter> */
		else if (!strcmp ((const char *)name, "bodymatter"))
		{
			ret = parseXMLBodymatter (reader, fragment, tmpNode);
			if (ret == 1) return ret;
			/*if (ret != 1) return ret; */
		}
		/* parse tag if <rearmatter> */
		else if (!strcmp ((const char *)name, "rearmatter"))
		{
			ret = parseXMLRearmatter (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <BOOK>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parse Frontmatter-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLFrontmatter (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;	
	int ret, type;
	ret = type = 1;
	
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <FRONTMATTER>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <FRONTMATTER>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	
	/* while we haven't reached the frontmatter-close tag */
	while (strcmp ((const char *)name, "frontmatter"))
	{	
		/* parse tag if <doctitle> */
		if (!strcmp ((const char *)name, "doctitle"))
		{
			ret = parseXMLDoctitle (reader);
			if (ret != 1) return ret;
		}
		/* parse tag if <docauthor> */
		else if (!strcmp ((const char *)name, "docauthor"))
		{
			ret = parseXMLDocauthor (reader);
			if (ret != 1) return ret;
		}
		/* parse tag if <covertitle> */
		else if (!strcmp ((const char *)name, "covertitle"))
		{
			ret = parseXMLCovertitle (reader);
			if (ret != 1) return ret;
		}
		/* parse tag if <level> */
		else if (!strcmp ((const char *)name, "level"))
		{
			ret = parseXMLLevel (reader);
			if (ret != 1) return ret;
		}
		/* parse tag if <levelx> */
		else if (!strcmp ((const char *)name, "level1") || !strcmp ((const char *)name, "level2") || !strcmp ((const char *)name, "level3")
		 || !strcmp ((const char *)name, "level4") || !strcmp ((const char *)name, "level5") || !strcmp ((const char *)name, "level6"))
		{
			ret = parseXMLLevel1 (reader, NULL, NULL);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <FRONTMATTER>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parse Bodymatter-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLBodymatter (xmlTextReaderPtr reader, char* fragment, struct Node *tmpNode)
{
	const xmlChar *name = NULL;	
	int ret, type;
	ret = type = 1;
	
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <BODYMATTER>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <BODYMATTER>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	
	/* while we haven't reached the bodymatter-close tag */
	while (strcmp ((const char *)name, "bodymatter"))
	{	
		/* parse tag if <level> */
		if (!strcmp ((const char *)name, "level"))
		{
			ret = parseXMLLevel (reader);
			if (ret != 1) return ret;
		}
		/* parse tag if <level1> */
		else if (!strcmp ((const char *)name, "level1"))
		{
			ret = parseXMLLevel1 (reader, fragment, tmpNode);
			if (ret == 1) return ret;
		/*	if (ret != 1) return ret; */
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <BODYMATTER>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parse Rearmatter-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLRearmatter (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;	
	int ret, type;
	ret = type = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <REARMATTER>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <REARMATTER>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	
	/* while we haven't reached the rearmatter-close tag */
	while (strcmp ((const char *)name, "rearmatter"))
	{	
		/* parse tag if <level> */
		if (!strcmp ((const char *)name, "level"))
		{
			ret = parseXMLLevel (reader);
			if (ret != 1) return ret;
		}
		/* parse tag if <level1> */
		else if (!strcmp ((const char *)name, "level1"))
		{
			ret = parseXMLLevel1 (reader, NULL, NULL);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <REARMATTER>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parse Doctitle-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLDoctitle (xmlTextReaderPtr reader)
{	
	const xmlChar *name = NULL;
	xmlChar *value = NULL;
	char * tmp = NULL;
	int ret = 1, type = 1;
	
	/* Attributes:*/
	if (DEBUG_NCXXML) report ("Parsing XML <DOCTITLE>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <DOCTITLE>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	/* if the title tag has a value	 */
	if (xmlTextReaderHasValue (reader)){
		value = xmlTextReaderValue (reader);
		tmp =  removewhitespaces ((char *)value);
		/* printf ("\t-%s-",tmp);
		  TODO: The title is never used.. */
	}
	if(value != NULL) 
		xmlFree(value); value = NULL;
	if(tmp != NULL)
		free(tmp); tmp = NULL;
	while (strcmp ((const char *)name, "doctitle"))
	{
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <DOCTITLE>, endtag expected", REP_ERROR);
		return -1;
	}
	return ret;
}


/**
 * Parse Docauthor-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLDocauthor (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	xmlChar *value = NULL;
	char *tmp = NULL;
	int ret =1 , type = 1;
	
	/* Attributes: */
	if (DEBUG_NCXXML) report ("Parsing XML <DOCAUTHOR>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <DOCAUTHOR>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	/* if the author tag has a value  */
	if (xmlTextReaderHasValue (reader)){
		value = xmlTextReaderValue (reader);
		tmp = removewhitespaces ((char *)value);
		/* printf ("\t-%s-",tmp);
		  TODO: The title is never used.. */
	}
	if(value != NULL) 
		xmlFree(value); value = NULL;
	if(tmp != NULL) free(tmp); tmp = NULL;
	while (strcmp ((const char *)name, "docauthor"))
	{
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <DOCAUTHOR>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parse Level-tag. Not implemented
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLLevel (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;	
	int ret = 1, type = 1;
	
	/* Attributes:
	 * depth
	 * class
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <LEVEL>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <LEVEL>", REP_ERROR);
		return -1;
	}
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	type = xmlTextReaderNodeType(reader);
	if (type != 15) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <LEVEL>, endtag expected", REP_ERROR);
		return -1;
	}	
	return ret;
}


/**
 * Parse Level1-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLLevel1 (xmlTextReaderPtr reader, char* fragment, struct Node *tmpNode)
{
	xmlChar *value = NULL, *tmp = NULL;
	const xmlChar *name = NULL;	
	int ret = 1, type = 1;

/*Level 1 may contain following childs..
  <level1>
    <pagenum>...</pagenum>
    <h1>...</h1>
    <level2>...</level2>
    <doctitle>...</doctitle>
    <docauthor>...</docauthor>
    <covertitle>...</covertitle>
    <bridgehead>...</bridgehead>
    <p>...</p>
    <list>...</list>
    <dl>...</dl>
    <div>...</div>
    <blockquote>...</blockquote>
    <img>...</img>
    <imggroup>...</imggroup>
    <poem>...</poem>
    <linegroup>...</linegroup>
    <byline>...</byline>
    <dateline>...</dateline>
    <epigraph>...</epigraph>
    <table>...</table>
    <address>...</address>
    <line>...</line>
    <author>...</author>
    <prodnote>...</prodnote>
    <sidebar>...</sidebar>
    <note>...</note>
    <a>...</a>
    <cite>...</cite>
    <samp>...</samp>
    <kbd>...</kbd>
</level1>
*/

	if (DEBUG_NCXXML) report ("Parsing XML <LEVEL1>", REP_DEBUG);
	type = xmlTextReaderNodeType(reader);
	if (type != 1) 
	{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <LEVEL1>", REP_ERROR);
		return -1;
	}
	
	/* if the fragment = NULL - we are in frontmatter */
	if (fragment == NULL)
	{	
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		if(xmlTextReaderHasValue (reader))
		{
			value = xmlTextReaderValue (reader);
			tmp =  (xmlChar *)removewhitespaces ((char *)value);
			/* printf ("\t-%s-",tmp);
			  TODO: The title is never used.. */
		}
	}
	/* we are in bodymatter */
	else
	{	
		/* check wheter it is the level1 tag who got the id we request */
		tmp = xmlTextReaderGetAttribute (reader, (xmlChar *)"id");
		ret = xmlTextReaderRead (reader);
		if (tmp != NULL)
		{	
			if (!strcmp ((char *)tmp, fragment))
			{
				while(!xmlTextReaderHasValue (reader))
				{
					ret = xmlTextReaderRead (reader);
					
					/* check wheter we have come to a new tag with id,
					 * if that is the case, the level1 tag didn't have
					 * any valid values 
					 */
					tmp = xmlTextReaderGetAttribute (reader, (xmlChar *)"id");
					if (tmp != NULL) return 0;
				}
				value = xmlTextReaderValue (reader);
				tmp =  (xmlChar *)removewhitespaces ((char *)value);
				if(tmp != NULL) tmpNode->textPassage = (char *)tmp;
				if(value != NULL) xmlFree(value); value = NULL;	
				/* printf ("\n%s\n",tmp); TODO store chatper title? */
				return 1;	/* return succesful reading	 */
			}
		}
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
		
		/* while we haven't reached the level1-endtag */
		while (strcmp ((const char *)name, "level1"))
		{	
			tmp = xmlTextReaderGetAttribute (reader, (xmlChar *)"id");
			if (tmp != NULL)
			{	
				/*
				name = xmlTextReaderConstName (reader);
				if (name == NULL) return 0;
				if (!strcmp (name, "p"))
				{
				*/
					if (!strcmp ((char *)tmp, fragment))
					{	ret = xmlTextReaderRead (reader);
						if(xmlTextReaderHasValue (reader))
						{
							value = xmlTextReaderValue (reader);
							tmp =  (xmlChar *)removewhitespaces ((char *)value);
							if(tmp != NULL) tmpNode->textPassage = (char *)tmp;
							if(value != NULL) xmlFree(value); value = NULL;										
							return 1;	
						}
						else
						{
							if(value != NULL) xmlFree(value); value = NULL;										
							if(tmp != NULL) xmlFree(tmp); tmp = NULL;																
							return 0;
						}
					}
				/*}*/				
			}
			ret = xmlTextReaderRead (reader);
			if (ret != 1) return ret;		
		}
		type = xmlTextReaderNodeType(reader);
		if (type != 15) 
		{   if (DEBUG_NCXPARSER) report ("Failed parsing XML <LEVEL1>, endtag expected", REP_ERROR);
			return -1;
		}			
	}
	if(value != NULL) 
		xmlFree(value); value = NULL;
	if(tmp != NULL)
		xmlFree(tmp); tmp = NULL;
	ret = xmlTextReaderRead (reader);
	return ret;
}


/**
 * Parse Covertitle-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLCovertitle (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;	
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <COVERTITLE>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;	
	return ret;
}

/*
 * Fetches the textvalue inside html-tags.
 * E.g: ....<b>text</b>...
 *
 * @param reader XmlTextReader.
 *
 * Return the textvalue.
 */
/*static char *parseXMLHtmltag (xmlTextReaderPtr reader)
{
	xmlChar *value = NULL;
	value = NULL;
	char *tempChar = NULL;
	
	xmlTextReaderRead (reader);
	TODO: hva om det ikke er flere noder etter denne?
	
	if (xmlTextReaderHasValue (reader))
	{
		value = xmlTextReaderValue (reader);
		tempChar = removewhitespaces(value);
		xmlFree (value);
	}
	return tempChar;	
}
*/


/************************* END of XML parser code *************************/


#endif
