/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file report.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 15.02.2006
 *
 * Functionality for handling error reporting.
 */
 

#include <stdio.h>
#include <stdlib.h>

#include "report.h"

/** sets the level for error reporting (0-4 where 0 is low and 4 high) */
int reporting_level = 4;
 
/**
 * Function for handling runtime message reporting to the user.
 * 
 * @param error_mesage - a string with the message to output.
 * @param level - the level of the message.
 */
void report (const char * error_message, const int level)
{
	switch (reporting_level)
	{
		case 4:
		 	if (level <= REP_DEBUG)		/* debug */
				printf ("Debug: %s\n", error_message);
	 	case 3:
		 	if (level == REP_MESSAGE)	/* message */
		 		printf ("Message: %s\n", error_message);
	 	case 2:
		 	if (level == REP_WARNING)	/* warning */
		 		printf ("Warning: %s\n", error_message);
	 	case 1:  
			if (level == REP_ERROR)		/* error */
				printf ("Error: %s\n", error_message);
		default:
			if (level >= REP_CRITICAL)	/* report critical error and exit program */
			{
				printf ("Critical Error: %s\n\nExiting program\n", error_message);
				exit(-1);
			}
	}
}
